from typing import Optional

from dask.dataframe import DataFrame
from returns.result import Result, Failure

from mason_dask.jobs.executed import ExecutedJob, InvalidJob
from pandas import DataFrame as PDataFrame

from fsspec.core import OpenFile, open_files
from pyexcelerate import Workbook
from dask import dataframe as dd, delayed
import pandas as pd

from mason_dask.utils.returns import safe_job


def df_to(df: DataFrame, output_path: str, output_format: str, label: Optional[str] = None) -> Result[ExecutedJob, InvalidJob]:
    
    def _write_excel(df: PDataFrame, fil: OpenFile):
        with fil as f:
            values = [df.columns] + list(df.values)
            wb = Workbook()
            wb.new_sheet('sheet 1', data=values)
            wb.save(f)
        return None



    if label:
        output_path = output_path + label + "/"

    success_message = f"Table succesfully formatted as {output_format} and exported to {output_path}"
    if output_format == "csv":
        @safe_job
        def to_csv(df, output_path):
            dd.to_csv(df, output_path, index=False)
            return success_message
        return to_csv(df, output_path)
    elif output_format == "parquet":
        @safe_job
        def to_parquet(df, output_path):
            dd.to_parquet(df, output_path)
            return success_message
        return to_parquet(df, output_path)
    elif output_format == "json":
        @safe_job
        def to_json(df, output_path):
            dd.to_json(df, output_path)
            return success_message
        return to_json(df, output_path) 
    elif output_format == "xlsx":
        @safe_job
        def to_xlsx(df: DataFrame, output_path: str):
            writer = pd.ExcelWriter('test_out.xlsx', engine='xlsxwriter')
            to_excel_chunk = delayed(_write_excel)

            dfs = df.repartition(partition_size="10MB").to_delayed()

            def name_function(i: int):
                return f"part_{i}.xlsx"

            files = open_files(
                output_path,
                mode="wb",
                num=df.npartitions,
                name_function=name_function
            )

            def replace_path(f: OpenFile) -> OpenFile:
                p = f.path
                f.path = p.replace(".xlsx.part", ".xlsx")
                return f

            files = list(map(lambda f: replace_path(f), files))
            values = [to_excel_chunk(dfs[0], files[0])]
            values.extend(
                [to_excel_chunk(d, f) for d, f in zip(dfs[1:], files[1:])]
            )
            delayed(values).compute()
        return to_xlsx(df, output_path)
    else:
        return Failure(InvalidJob(f"Invalid output format: {output_format}"))





from os import path, listdir
from shutil import rmtree

import pytest
from distributed import Client, LocalCluster

from mason_dask.definitions import from_root
from mason_dask.jobs.executed import InvalidJob
from mason_dask.jobs.format import FormatJob, ValidFormatJob
from dask import dataframe as dd

from mason_dask.utils.cluster_spec import ClusterSpec

TMP = from_root("/.tmp/")
TEST_DATA = from_root("/test/data/test_data.csv")

@pytest.fixture(scope="module")
def client():
    with Client(LocalCluster(n_workers=2)) as client:
        yield(client)  

@pytest.fixture(autouse=True)
def run_around_tests():
    yield
    clear_tmp()
        
def clear_tmp():
    if path.exists(TMP):
        rmtree(TMP)

def test_local_client(client):
    spec = {
        "input_paths": ["s3://mason-sample-data/tests/in/csv/sample2.csv"],
        "input_format": "csv",
        "output_format": "csv",
        "output_path": "s3://mason-sample-data/tests/out/csv/",
        "filter_columns": [],
        "partition_columns": [],
        "line_terminator": "\r\n",
        "partitions": "3"
    }

    job = FormatJob(spec).validate()
    assert(isinstance(job, ValidFormatJob))

    cluster_spec = ClusterSpec(client)
    assert(cluster_spec.worker_specs is not None)
    assert (len(cluster_spec.worker_specs) == 2)

def test_schema(client):
    in_paths = ["good_input_path", "good_input_path2"]
    in_format = "csv"
    out_format = "parquet"
    out_path = "good_output_path"
    filter_columns = ["col1", "col2"]
    partition_columns = ["col3", "col4"]

    good_spec = {
        "input_paths": in_paths,
        "input_format": in_format,
        "output_format": out_format,
        "output_path": out_path,
        "filter_columns": filter_columns,
        "partition_columns": partition_columns 
    }
    
    job = FormatJob(good_spec).validate()
    assert(isinstance(job, ValidFormatJob))
    assert(job.input_paths == in_paths)
    assert(job.input_format == in_format)
    assert(job.output_format == out_format)
    assert(job.line_terminator == "\n")
    assert(job.partitions == None)
    assert(job.output_path == out_path + "/")
    assert(job.filter_columns == filter_columns)
    assert(job.partition_columns == partition_columns)

    bad_spec = {
        "input_paths": "1234",
        "input_format": "avro",
        "output_format": "",
        "line_terminator": "",
        "output_path": "1234",
        "filter_columns": "bad stuff",
        "partition_columns": "bad stuff"
    }

    for key, value in bad_spec.items():
        good_spec[key] = value
        job = FormatJob(good_spec).validate()
        assert(isinstance(job, InvalidJob))

def test_csv(client):
    # baseline test
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_format": "csv",
        "output_path": TMP + "csv/",
        "filter_columns": [],
        "partition_columns": [],
    }


    cluster_spec = ClusterSpec(client)
    assert(cluster_spec.worker_specs is not None)
    assert(len(cluster_spec.worker_specs) == 2)

    job = FormatJob(spec).validate()
    assert(isinstance(job, ValidFormatJob))
    result = job.run(cluster_spec)

    parts = ["0.part", '1.part']
    assert(sorted(listdir(TMP + "csv/")) == parts)
    for p in parts:
        df = dd.read_csv(TMP + f"/csv/{p}").compute()
        assert(df.shape[0] == 2)
        assert(sorted(list(df["col_a"])) == [123.0, 789.0])

    clear_tmp()

    # one partition column
    spec["partition_columns"] = ["col_a"]
    job = FormatJob(spec).validate()
    assert(isinstance(job, ValidFormatJob))
    job.run(cluster_spec)

    prts = [789.0, 123.0]
    assert(listdir(TMP + "csv/") == list(map(lambda p: f"col_a={p}", prts)))

    for prt in prts:
        df = dd.read_csv(TMP + f"/csv/col_a={prt}/0.part").compute()
        assert(df.shape[0] == 2)
        assert(sorted(list(df["col_a"])) == [prt, prt])

    # two partition columns
    clear_tmp()
    spec["partition_columns"] = ["col_a", "col_b"]
    job = FormatJob(spec).validate()
    assert(isinstance(job, ValidFormatJob))
    job.run(cluster_spec)

    part_dict = {(789.0, "test3"): [789.0], (789.0, "test4"): [789.0], (123.0, "test2"): [123.0,123.0]}
    assert(sorted(listdir(TMP + "csv/")) == sorted(list(map(lambda p: f"col_a={p[0]}&col_b={p[1]}", list(part_dict.keys())))))

    for item in part_dict.items():
        df = dd.read_csv(TMP + f"/csv/col_a={item[0][0]}&col_b={item[0][1]}/0.part").compute()
        assert(df.shape[0] == len(item[1]))
        assert(sorted(list(df["col_a"])) == item[1]) 

    # filter columns
    clear_tmp()
    spec["partition_columns"] = []
    spec["filter_columns"] = ["col_a", "col_b"]

    job = FormatJob(spec).validate()
    assert (isinstance(job, ValidFormatJob))
    job.run(cluster_spec)

    parts = ["0.part", '1.part']
    assert(sorted(listdir(TMP + "/csv/")) == parts)
    df = dd.read_csv(TMP + f"/csv/*").compute()
    assert(df.shape[0] == 4)
    assert(list(df.columns) == ["col_a", "col_b"])

    # partitions options
    spec["filter_columns"] = []
    spec["partitions"] = "4"

    job = FormatJob(spec).validate()
    assert (isinstance(job, ValidFormatJob))
    job.run(cluster_spec)
    assert(sorted(listdir(TMP + "csv/")) == ["0.part", "1.part", "2.part", "3.part"])

def test_parquet(client):
    # baseline
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_format": "parquet",
        "output_path": TMP + "parquet/",
        "filter_columns": [],
        "partition_columns": [],
    }

    job = FormatJob(spec).validate()
    assert (isinstance(job, ValidFormatJob))
        
    cluster_spec = ClusterSpec(client)

    job.run(cluster_spec)

    parts = ["part.0.parquet", 'part.1.parquet']
    assert(sorted(list(filter(lambda f: f.endswith(".parquet"),listdir(TMP + "parquet/")))) == parts)
    for p in parts:
        df = dd.read_parquet(TMP + f"/parquet/{p}").compute()
        assert(df.shape[0] == 2)
        assert(sorted(list(df["col_a"])) == [123.0, 789.0])

    # partition_columns
    spec["partition_columns"] = ["col_a"]

    job = FormatJob(spec).validate()
    assert(isinstance(job, ValidFormatJob))
    job.run(cluster_spec)

    prts = [789.0, 123.0]
    assert(sorted(list(filter(lambda f: f.endswith(".parquet"), listdir(TMP + "parquet/")))) == ["part.0.parquet", "part.1.parquet"])

    for prt in prts:
        df = dd.read_parquet(TMP + f"parquet/col_a={prt}/part.0.parquet").compute()
        assert(df.shape[0] == 2)
        assert(sorted(list(df["col_a"])) == [prt, prt])


def test_json(client):
    # baseline
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_format": "json",
        "output_path": TMP + "json/",
        "filter_columns": [],
        "partition_columns": [],
    }

    job = FormatJob(spec).validate()
    assert (isinstance(job, ValidFormatJob))

    cluster_spec = ClusterSpec(client)

    job.run(cluster_spec)
    
    parts = ["0.part", '1.part']
    assert(sorted(listdir(TMP + "/json/")) == parts)
    for p in parts:
        df = dd.read_json(TMP + f"/json/{p}").compute()
        assert(df.shape[0] == 2)
        assert(sorted(list(df["col_a"])) == [123, 789])
        
def test_xlsx(client):
    # baseline
    spec = {
        "input_paths": [TEST_DATA],
        "input_format": "csv",
        "output_format": "xlsx",
        "output_path": TMP + "xlsx/",
        "filter_columns": [],
        "partition_columns": [],
    }

    job = FormatJob(spec).validate()
    assert (isinstance(job, ValidFormatJob))
    with Client(LocalCluster(n_workers=2)) as client:
        cluster_spec = ClusterSpec(client)
        job.run(cluster_spec)

        parts = ["part_0.xlsx"]
        assert(sorted(listdir(TMP + "xlsx/")) == parts)





from collections import namedtuple
from typing import Union, List

from dask.dataframe import DataFrame
from mason_dask.utils.cluster_spec import ClusterSpec
from returns.result import Result, Success

from mason_dask.jobs.executed import InvalidJob, ExecutedJob
from mason_dask.jobs.format import VALID_INPUT_FORMATS
from schema import Schema, And, Use, SchemaError
from schema import Optional as SOptional

from mason_dask.models.dataframe_from import df_from
from mason_dask.models.dataframe_to import df_to as df_to_output
from mason_dask.utils.exception import message

from returns.pipeline import flow
from returns.pointfree import bind

from dask_sql import Context


class ValidQueryJob:

    def __init__(self, t):
        self.input_format: str = t.input_format
        self.input_paths: List[str] = t.input_paths
        self.output_path: str = t.output_path
        self.query_string: str = t.query_string
        self.line_terminator: str = t.line_terminator
        self.table_name: str = t.table_name

    def run(self, cluster_spec: ClusterSpec) -> Result[ExecutedJob, InvalidJob]:
        return flow(
            self.df(),
            bind(self.query),
            bind(self.df_to)
        )
        
    def df(self) -> Result[DataFrame, InvalidJob]:
        return df_from(self.input_paths, self.input_format, self.line_terminator)
    
    def query(self, dataframe: DataFrame) -> Result[DataFrame, InvalidJob]:
        c = Context()
        c.register_dask_table(dataframe, self.table_name)
        queried = c.sql(self.query_string)
        return Success(queried)
    
    def df_to(self, dataframe: DataFrame) -> Result[ExecutedJob, InvalidJob]:
        return df_to_output(dataframe, self.output_path, "parquet")

class QueryJob:
    def __init__(self, spec: dict):
        self.spec = spec

    def schema(self) -> Schema:
        input_formats = And(Use(str), lambda n: n in VALID_INPUT_FORMATS)

        schema = {
            "input_format": input_formats,
            "input_paths": [Use(str)],
            "output_path": Use(str),
            "query_string": Use(str),
            SOptional("line_terminator", default="\n"): str,
            SOptional("table_name", default="$table"): str
        }
        return Schema(schema)

    def validate(self) -> Union[ValidQueryJob, InvalidJob]:
        try:
            d = self.schema().validate(self.spec)
            t = namedtuple("JobAttributes", d.keys())(*d.values())
            return ValidQueryJob(t)
        except SchemaError as e:
            return InvalidJob(f"Invalid Schema: {message(e)}")






from typing import List, Tuple, Optional

from distributed import Client
from distributed.scheduler import WorkerState

from mason_dask.utils.worker_spec import WorkerSpec

class ClusterSpec:
    
    def __init__(self, client: Client, num_workers: Optional[int] = None, desired_spec: Optional[WorkerSpec] = None, scheduler: Optional[str] = None):
        self.worker_specs = None
        if client.cluster and client.cluster.scheduler:
            workers_dict: dict = client.cluster.scheduler.workers
            workers: List[WorkerState] = list({k: workers_dict[k] for k in workers_dict.keys() if isinstance(workers_dict[k], WorkerState)}.values())
            worker_specs: List[Tuple[int, int]] = list(map(lambda w: (w.memory_limit, w.ncores), workers))
            self.worker_specs = list(map(lambda s: WorkerSpec(*s), worker_specs))
        elif num_workers and desired_spec:
            assert(desired_spec is not None)
            assert(num_workers is not None)
            self.worker_specs = [desired_spec] * num_workers 
        self.scheduler = scheduler
            
    def valid(self) -> bool:
        if self.worker_specs:
            return True
        else:
            return False

    def max_partition_size(self) -> Optional[int]:
        if self.worker_specs:
            return min(list(map(lambda s: s.memory, self.worker_specs)))
        else:
            return None
    
    def num_workers(self) -> Optional[int]:
        if self.worker_specs:
            return len(self.worker_specs)
        else:
            return None
    



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.readlines()

with open('VERSION') as f:
    version = f.read()

long_description = 'Mason Dask Exeuction Jobs'

setup(
    name='mason-dask',
    version=version,
    author='Kyle Prifogle',
    author_email='kyle.prifogle@samtec.com',
    url='',
    description='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='',
    packages=find_packages(),
    install_requires=requirements,
    zip_safe=False,
    include_package_data=True,
    python_requires='>=3.6',
)

